;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 Abersoft fig-FORTH recompiled
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(*
registers:
  BC -- IP
  SP -- data stack
  IX -- NEXT address (why, it is faster this way!)

  for asm code i used disassembly done by Marcos Cruz (programandala.net)
  http://programandala.net/en.program.abersoft_forth.html

  original code by John Jones-Steel, 1983
  heavily modified (hugely rewritten) by Ketmar Dark
*)


$use <zxasm>

enum{
  def: ITC
  def: DTC
  def: STC
}

;; threaded code type
DTC quan OPT-TC-TYPE

: OPT-ITC?  ( -- flag )  OPT-TC-TYPE ITC = ;
: OPT-DTC?  ( -- flag )  OPT-TC-TYPE DTC = ;
: OPT-STC?  ( -- flag )  OPT-TC-TYPE STC = ;


enum{
  def: R/W-TAPE
  def: R/W-ZXEMUT
  def: R/W-P3DOS
}

;; which kind of block i/o to use?
R/W-ZXEMUT quan OPT-BLOCK-I/O-TYPE


;; relative BRANCHes?
;; non-relative branches are slightly faster
;; (and the code is smaller by almost 20 bytes)
false quan OPT-REL-BRANCH?

;; include "CLIT" byte literals?
;; the code is very slightly smaller this way.
true quan OPT-CLIT?

;; include "BLIT" byte literals?
;; almost no saved bytes here.
false quan OPT-BLIT?

;; include various screen words?
false quan OPT-SCREEN-WORDS?

;; include various additional screen words?
false quan OPT-MORE-SCREEN-WORDS?

;; include line editor code?
false quan OPT-EDITOR-CODE?

;; include fullscreen editor code?
false quan OPT-FS-EDITOR-CODE?

;; print driver to use (# of columns)
32 quan OPT-EMIT-DRIVER?

;; use 64 chars line in 64-cols fullscreen editor?
false quan OPT-FS4-64?

;; use ROM font for small letters?
;; this saves 234 bytes.
true quan OPT-EMIT6-UGLY?

;; include 16-bit unsigned "UU*" and "UU/MOD"?
true quan OPT-16BIT-MUL/DIV?

;; use unrolled code to gain more speed?
;; costs ~108 bytes
true quan OPT-16BIT-MUL/DIV-UNROLLED?

;; include shift words?
true quan OPT-SHIFT-WORDS?

;; new word header format?
true quan OPT-NEW-ART-HEADER?

;; use simple byte-sized hash before comparing a name?
;; demo game loading is ~1 second faster with it, and it wasted ~500 bytes.
;; totally doesn't worth it.
false quan OPT-NEW-ART-HEADER-HASHED?

;; move system and compiler support words to "SYSTEM" vocabulary?
false quan OPT-SYS-VOCAB?


;; NOT use simpler, but fig-compatible BEGIN/IF/etc?
true quan OPT-ADVANCED-CFLOW?

;; include Parnas' generalised flow contol words?
false quan OPT-PARNAS?


;; include floating-point library?
false quan OPT-ENABLE-FP?

;; include "UNDER+" and others?
true quan OPT-UNDER-MATH?

;; setup custom interrupt handler?
true quan OPT-IM2-HANDLER?


;; include "SPRX" library? ~420 bytes
false quan OPT-SPRX-LIB?

;; including both engines costs ~440 bytes
;; including 3 engines costs ~625 bytes

;; include "XOR BLIT"? ~250 bytes
true quan OPT-SPRX-XOR-MODE?

;; include "OR BLIT"? ~250 bytes
false quan OPT-SPRX-OR-MODE?

;; include "AND BLIT"? ~250 bytes
false quan OPT-SPRX-AND-MODE?

;; include SPX demo sprites?
false quan OPT-DEMO-SPRITES?


;; include "SPR16" library?
false quan OPT-SPR16-LIB?


;; include ZXEmuT debug printing?
false quan OPT-ZXEMUT-DPRINT?


;; include "eratos test"?
false quan OPT-ETEST?

;; save demo disc from "demo/game.blk"?
false quan OPT-SAVE-DEMO-DISC?

;; verbose "LOAD"?
false quan OPT-VERBOSE-LOAD?

;; turn on ZXEmuT "maxspeed" while loading?
false quan OPT-FAST-LOAD?


true not quan OPT-ROOM-EDITOR?
true quan OPT-SPRITE-EDITOR?


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CLI arguments parsing

: OPT-32-COLS?  OPT-EMIT-DRIVER? 32 = ;
: OPT-42-COLS?  OPT-EMIT-DRIVER? 42 = ;
: OPT-64-COLS?  OPT-EMIT-DRIVER? 64 = ;


$use <getarg>

0 quan output-file-name ;; dynalloced

|: set-output-fname  ( addr count )
  output-file-name ?error" too many output files" string:$new output-file-name:! ;

getarg:arg: --output  " file name?" getarg:next-arg set-output-fname ;
getarg:last-need-arg
getarg:help: ." output tape file (default is \'tap/aforth.tap\')" ;
getarg:last-alias: -o

getarg:arg: --dtc DTC OPT-TC-TYPE:! ;
OPT-DTC? [IFNOT] getarg:help: ." use Direct Threaded Code" ; [ENDIF]
getarg:arg: --itc ITC OPT-TC-TYPE:! ;
OPT-ITC? [IF] getarg:help: ." use Indirect Threaded Code" ; [ENDIF]
\ getarg:arg: --stc STC OPT-TC-TYPE:! ;
\ OPT-STC? [IF] getarg:help: ." use Subroutine Threaded Code" ; [ENDIF]

getarg:arg: --r/w-tape R/W-TAPE OPT-BLOCK-I/O-TYPE:! ;
OPT-BLOCK-I/O-TYPE R/W-TAPE = [IFNOT] getarg:help: ." use tape block i/o" ; [ENDIF]
getarg:arg: --r/w-zxemut R/W-ZXEMUT OPT-BLOCK-I/O-TYPE:! ;
OPT-BLOCK-I/O-TYPE R/W-ZXEMUT = [IFNOT] getarg:help: ." use ZXEmuT traps block i/o" ; [ENDIF]
getarg:arg: --r/w-p3dos R/W-P3DOS OPT-BLOCK-I/O-TYPE:! ;
OPT-BLOCK-I/O-TYPE R/W-P3DOS = [IFNOT] getarg:help: ." use +3DOS block i/o" ; [ENDIF]


getarg:arg: --rel-branches OPT-REL-BRANCH?:!t ;
OPT-REL-BRANCH? [IFNOT] getarg:help: ." use relative branch offsets" ; [ENDIF]
getarg:arg: --abs-branches OPT-REL-BRANCH?:!f ;
OPT-REL-BRANCH? [IF] getarg:help: ." use absolute branch offsets" ; [ENDIF]

getarg:arg: --32 32 OPT-EMIT-DRIVER?:! ;
OPT-32-COLS? [IFNOT] getarg:help: ." use 32 columns print driver" ; [ENDIF]
getarg:arg: --42 42 OPT-EMIT-DRIVER?:! ;
OPT-42-COLS? [IFNOT] getarg:help: ." use 42 columns print driver" ; [ENDIF]
getarg:arg: --64 64 OPT-EMIT-DRIVER?:! ;
OPT-64-COLS? [IFNOT] getarg:help: ." use 64 columns print driver" ; [ENDIF]

getarg:arg: --emit6-ugly OPT-EMIT6-UGLY?:!t ;
OPT-EMIT6-UGLY? [IFNOT] getarg:help: ." use ugly, but smaller 6x8 font (saves 234 bytes)" ; [ENDIF]
getarg:arg: --emit6-nice OPT-EMIT6-UGLY?:!f ;
OPT-EMIT6-UGLY? [IF] getarg:help: ." use nice, but bigger 6x8 font" ; [ENDIF]

" Parnas' flow control" getarg:enable/disable-arg: OPT-PARNAS? parnas
" CLIT byte-sized literal primitive" getarg:enable/disable-arg: OPT-CLIT? clit
" BLIT byte-sized literal primitive" getarg:enable/disable-arg: OPT-BLIT? blit
" bit shifting words" getarg:enable/disable-arg: OPT-SHIFT-WORDS? shift-words
" \'UNDER+\' and company" getarg:enable/disable-arg: OPT-UNDER-MATH? under-math
" basic screen words" getarg:enable/disable-arg: OPT-SCREEN-WORDS? extra-screen-words
" rarely used screen words" getarg:enable/disable-arg: OPT-MORE-SCREEN-WORDS? basic-screen-words
" line editor words" getarg:enable/disable-arg: OPT-EDITOR-CODE? line-editor
" fullscreen editor words" getarg:enable/disable-arg: OPT-FS-EDITOR-CODE? fs-editor
" ROM floating point library" getarg:enable/disable-arg: OPT-FS-EDITOR-CODE? rom-fp
" SYSTEM vocabulary" getarg:enable/disable-arg: OPT-SYS-VOCAB? sys-vocab
" IM2 handler" getarg:enable/disable-arg: OPT-IM2-HANDLER? im2

" simple LB-like sprite engine" getarg:enable/disable-arg: OPT-SPRX-LIB? sprx
" simple LB-like sprite engine XOR mode" getarg:enable/disable-arg: OPT-SPRX-XOR-MODE? sprx-xor
" simple LB-like sprite engine AND mode" getarg:enable/disable-arg: OPT-SPRX-AND-MODE? sprx-and
" simple LB-like sprite engine OR mode" getarg:enable/disable-arg: OPT-SPRX-OR-MODE? sprx-or

" simple 16x16 XOR sprite engine" getarg:enable/disable-arg: OPT-SPR16-LIB? spr16-xor

;; hidden options
0 getarg:enable/disable-arg: OPT-ETEST? etest
0 getarg:enable/disable-arg: OPT-ZXEMUT-DPRINT? zxemut-dprint
0 getarg:enable/disable-arg: OPT-SAVE-DEMO-DISC? demo-disk
0 getarg:enable/disable-arg: OPT-VERBOSE-LOAD? verbose-load
0 getarg:enable/disable-arg: OPT-FAST-LOAD? fast-load
0 getarg:enable/disable-arg: OPT-NEW-ART-HEADER? new-headers
0 getarg:enable/disable-arg: OPT-ADVANCED-CFLOW? advanced-cflow

0 getarg:enable/disable-arg: OPT-DEMO-SPRITES? demo-sprite-data


getarg:usage:  ." urforth aber-main.f [args]\nargs:\n" ;
getarg:parse-args

output-file-name [IFNOT]
" tap/aforth.tap" set-output-fname
[ENDIF]


OPT-EMIT-DRIVER? dup 32 =
over 42 = or
swap 64 = or " invalid print driver config" not?error

\ OPT-FASTER-MUL/DIV? [IF]
\   OPT-16BIT-MUL/DIV? " cannot use faster mul/div without fast 16-bit unsigned mul/div" not?error
\ [ENDIF]

OPT-FS4-64? [IF]
  OPT-EMIT4? [IFNOT] OPT-FS4-64?:!0 [ENDIF]
[ENDIF]


: on/off.  ( flag )  ?< ." en" || ." dis" >? ." abled\n" ;

: opt.  ( addr count flag )  endcr >r type ." : " r> on/off. ;

: options.
  endcr
  OPT-FS-EDITOR-CODE? OPT-EDITOR-CODE? or ?<
    OPT-FS-EDITOR-CODE? OPT-EDITOR-CODE? xor
    not?error" use either line, or fullscreen editor. not both!" >?
  \ OPT-FS-EDITOR-CODE? ?<
  \   OPT-EMIT6? not?< (sp0!) error" cannot use fullscreen editor with 32-column driver" >? >?
  ." build options:\n"
  ."   threaded code type: " OPT-TC-TYPE <<
    ITC of?v| ." ITC\n" |?
    DTC of?v| ." DTC\n" |?
    STC of?v| ." STC\n" |?
    else| abort >>
  ."   block i/o: " OPT-BLOCK-I/O-TYPE <<
    R/W-TAPE of?v| ." tape\n" |?
    R/W-ZXEMUT of?v| ." ZXEmuT traps\n" |?
    R/W-P3DOS of?v| ." +3DOS\n" |?
    else| abort >>
  ."   header format: " OPT-NEW-ART-HEADER? ?< ." new\n" || ." old\n" >?
  ."   branch offsets: " OPT-REL-BRANCH? ?< ." relative\n" || ." absolute\n" >?
  ."   print driver: " OPT-EMIT-DRIVER? . ." columns\n"
   "   CLIT" OPT-CLIT? opt.
   "   BLIT" OPT-BLIT? opt.
   "   under math" OPT-UNDER-MATH? opt.
  ."   16-bit mul/div: " OPT-16BIT-MUL/DIV? ?<
    OPT-16BIT-MUL/DIV-UNROLLED? ?< ." unrolled\n" || ." faster\n" >?
   || ." disabled\n" >?
   "   bit shifts" OPT-SHIFT-WORDS? opt.
   "   basic screen words" OPT-SCREEN-WORDS? opt.
   "   extra screen words" OPT-MORE-SCREEN-WORDS? opt.
   "   editor words" OPT-EDITOR-CODE? opt.
   "   fullscreen editor" OPT-FS-EDITOR-CODE? opt.
   "   floating point lib" OPT-ENABLE-FP? opt.
   "   custom IM2 handler" OPT-IM2-HANDLER? opt.
;
options.


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; main code

;; define some constants
module TCOM
;; base address for the new system
\ $6000 quan zx-base-addr
$8000 quan zx-base-addr


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; calculate disk buffer sizes

;; i/o buffer size. other values are calculated.
;; WARNING! must be power of 2!
( 128) 256 constant zx-i/o-buf-size
zx-i/o-buf-size pot 1+ " invalid ZX buffer size" not?error
zx-i/o-buf-size 64 mod " invalid ZX buffer size" ?error
1024 zx-i/o-buf-size /mod " invalid ZX I/O buffer size" ?error constant zx-b/scr
;; 4 extra bytes are used internally.
zx-i/o-buf-size 4 + constant zx-i/o-buf-full-size


OPT-BLOCK-I/O-TYPE R/W-TAPE = [IF]
;; leave room for RAM-DISC
$CB40 constant zx-s0
$CBE0 constant zx-r0

;; Start of RAM-DISC area.
$D000 constant zx-ramdisc^
;; RAM-DISC size.
$2C00 constant zx-#ramdisc

zx-ramdisc^ constant zx-buffer-lo
zx-ramdisc^ zx-#ramdisc + 1- constant zx-buffer-hi

[ELSE]
;; UDG at $FF58
;; we need bytes from $FFF4 to setup IM2 handler, so last 2 UDG chars are unusable.

$FFF0 constant zx-mem-bottom

;; will be slightly more.
;; this gives us 4 buffers of 256 bytes, exactly for one 1024 byte block.
1024 constant zx-rough-buffer-area-size

;; Start of RAM-DISC area.
zx-mem-bottom constant zx-ramdisc^

;; no need to have any RAM-DISC
zx-mem-bottom  zx-rough-buffer-area-size zx-i/o-buf-size / zx-i/o-buf-full-size * -
$FFF0 forth:and

dup constant zx-r0
160 - constant zx-s0
[ENDIF]
endcr ." ZX Forth S0: $" zx-s0 .hex4 cr
endcr ." ZX Forth R0: $" zx-r0 .hex4 cr

zx-s0 constant zx-query-tib

;; A constant that leaves the address of the first (lowest) block buffer.
zx-r0 constant zx-buffer-first

;; A constant leaving the address just above the highest memory address used
;; by the system.
zx-ramdisc^ zx-buffer-first - zx-i/o-buf-full-size /
zx-i/o-buf-full-size * zx-buffer-first + constant zx-buffer-limit

zx-buffer-limit zx-buffer-first - zx-i/o-buf-full-size / constant zx-i/o-buf-count
zx-b/scr pot 1+ " invalid ZX B/SCR" not?error
zx-i/o-buf-count 2 < " need at least 2 ZX buffers" ?error

1 [IF]
endcr ." ZX I/O BUFFER FIRST: $" zx-buffer-first .hex4 cr
endcr ." ZX I/O BUFFER LIMIT: $" zx-buffer-limit .hex4 cr
endcr ." ZX I/O BUFFER COUNT: " zx-i/o-buf-count 0 .r, cr
endcr ." ZX I/O BUFFER: " zx-i/o-buf-size ., ." bytes.\n"
\ endcr ." ZX I/O BUFFER FULL: " zx-i/o-buf-full-size ., ." bytes.\n"
endcr ." ZX I/O B/SCR : " zx-b/scr 0 .r, cr
[ENDIF]
end-module


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; load target compiler

$include "tcom/tcom-10-base.f"
$include "tcom/tcom-20-creatori.f"
$include "tcom/tcom-30-if-begin.f"
$include "tcom/tcom-36-parnas.f"


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; load ZX Forth system sources

$include "figsrc/zx-10-startup.f"
tcom:create-first-zx-word

tcom:zx-vocabulary EDITOR tcom:zx-immediate
tcom:zx-latest-vocid tcom:zx-editor-vocid:!

OPT-SYS-VOCAB? [IF]
tcom:zx-vocabulary (SYS) tcom:zx-immediate
tcom:zx-latest-vocid tcom:zx-system-vocid:!
[ENDIF]

$include "figsrc/zx-20-prims.f"
$include "figsrc/zx-22-prims-math.f"
$include "figsrc/zx-24-prims-muldiv.f"
$include "figsrc/zx-26-prims-mem.f"
$include "figsrc/zx-28-prims-io.f"
$include "figsrc/zx-30-low-level.f"


;; it is here strictly to maintain search order
OPT-ENABLE-FP? [IF]
$include "figsrc/zx-80-libs-10-float.f"
[ENDIF]

$include "figsrc/zx-32-blocks-and-more.f"
$include "figsrc/zx-34-if-then.f"
OPT-PARNAS? [IF]
$include "figsrc/zx-36-parnas.f"
[ENDIF]
$include "figsrc/zx-40-creatori.f"
$include "figsrc/zx-50-interpret.f"
$include "figsrc/zx-60-other.f"

$include "tcom/tcom-80-see.f"

OPT-EDITOR-CODE? [IF]
$include "figsrc/zx-70-editor.f"
[ENDIF]
OPT-FS-EDITOR-CODE? [IF]
$include "figsrc/zx-76-fs-editor.f"
[ENDIF]

;; input scanner ("stick reading")
$include "figsrc/zx-80-libs-14-stick-scan.f"


OPT-SPRX-LIB? [IF]
;; demo sprites
OPT-DEMO-SPRITES? [IF]
extend-module TCOM

<zx-asm>
zx-demo-sprites:
<end-asm>

" misc/sprites.wls" zx-incbin 0 zx-w,
@asm-label: zx-demo-sprites zx-constant DEMO-SPRITES

end-module \ TCOM
[ENDIF]
;; sprite library
$include "figsrc/zx-80-libs-20-sprx.f"
[ENDIF]


OPT-SPR16-LIB? [IF]
$include "figsrc/zx-80-libs-30-spr16.f"
[ENDIF]


OPT-ETEST? [IF]
$include "figsrc/zx-96-etest.f"
[ENDIF]


OPT-SPRITE-EDITOR? [IF]
$include "figsrc/zx-80-libs-40-arrow-mouse.f"
$include "figsrc/zx-90-apps-10-spr16-edit.f"
[ENDIF]

OPT-ROOM-EDITOR? [IF]
$include "figsrc/zx-80-libs-40-arrow-mouse.f"
$include "figsrc/zx-90-apps-20-room-edit.f"
[ENDIF]


vsp-depth " vsp fucked" ?error
nsp-depth " nsp fucked" ?error

\ tcom:zx-see VOCABULARY
\ tcom:zx-see definitions
\ tcom:zx-see [
\ tcom:zx-see abort
\ tcom:zx-see interpret
\ tcom:zx-see quit
\ tcom:zx-see expect
\ tcom:zx-see key
\ tcom:zx-see type
\ tcom:zx-see NUM-CHECK-PFX
\ tcom:zx-see NUMBER

tcom:check-forwards
zxa:finalise

tcom:zx-dp^ tcom:zx-w@ tcom:zx-fence^ tcom:zx-w!

;; DP, FENCE, VOCLINK
tcom:zx-dp^ tcom:zx-w@ tcom:zx-dpx-init^ tcom:zx-w!
tcom:zx-dp^ 2 + tcom:zx-w@ tcom:zx-dpx-init^ 2 + tcom:zx-w!
tcom:zx-dp^ 4 + tcom:zx-w@ tcom:zx-dpx-init^ 4 + tcom:zx-w!
;; FORTH-LATEST
tcom:zx-forth-vocid tcom:zx-w@ tcom:zx-dpx-init^ 6 + tcom:zx-w!
;; EDITOR-LATEST
tcom:zx-editor-vocid tcom:zx-w@ tcom:zx-dpx-init^ 8 + tcom:zx-w!

\ tcom:zx-see COLD
\ tcom:zx-see SPACES

\ tcom:zx-see (.INK)

\ tcom:zx-see x00
\ tcom:zx-see boo
\ tcom:zx-see goo
\ tcom:zx-see ?SPACE

\ tcom:zx-see interpret

\ tcom:zx-see wdemit


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; show statistics

extend-module TCOM

endcr
zx-here zx-base-addr -
." total system size: " .bytes cr
." system start address: $" zx-base-addr .hex4 cr
."   system end address: $" zx-here .hex4 cr
." statistics:\n"
."   total word count: " zx-stats-words 0.r, cr
."   header count: " zx-stats-headers 0.r, cr
."   headers size: " zx-stat-header-bytes .bytes cr
."   CFAs size   : " zx-stat-cfa-bytes .bytes cr
."   code words : " zx-stats-code 0.r, cr
."   mcode size : " zx-stat-mcode-bytes .bytes cr
."   colon words: " zx-stats-colon 0.r, cr
."   tcode size : " zx-stat-tcode-bytes .bytes cr
."   variable words: " zx-stats-var 0.r, cr
."   constant words: " zx-stats-const 0.r, cr
."   uservar words : " zx-stats-uvar 0.r, cr
."   quan words    : " zx-stats-quan 0.r, cr
."   vect words    : " zx-stats-vect 0.r, cr
."   alias words   : " zx-stats-alias 0.r, cr
zx-stat-clit-bytes-saved [IF]
."   saved by CLIT: " zx-stat-clit-bytes-saved .bytes cr
[ENDIF]
zx-stat-blit-bytes-saved [IF]
."   saved by BLIT: " zx-stat-blit-bytes-saved .bytes cr
[ENDIF]
."   forward references: " zx-stat-forwards 0.r, cr

zx-here zx-s0 128 - u>= " system image is too big" ?error
end-module


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; write tape

output-file-name count
string:extract-ext " .dsk" string:=ci [IF]
  ;; save disk
  endcr ." SAVING DSK: " output-file-name count type cr

  \ mk-cpcdsk:p3-data mk-cpcdsk:new-dsk
  mk-cpcdsk:pcw-ds mk-cpcdsk:new-dsk
  zxa:dsk:set-autoload-cargador-name
  zxa:dsk:save-all

  1 [IF]
  " ROOMEDIT.BLK" mk-cpcdsk:create-file
  " _zxemut_.DISC.blk" file:open-r/o quan blk-fd
  here  blk-fd file:size  blk-fd file:read-exact
  blk-fd file:size blk-fd file:close
  here swap mk-cpcdsk:write
  mk-cpcdsk:close-file
  [ENDIF]

  output-file-name count file:create
  dup mk-cpcdsk:save-dsk file:close

  mk-cpcdsk:free-dsk
[ELSE]
  ;; save tape
  " AFORTH" zxa:tap:set-cargador-name
  output-file-name count
  endcr ." SAVING TAPE: " 2dup type cr
  file:create quan tap-fd
  tap-fd zxa:tap:save-fd
  OPT-SAVE-DEMO-DISC? [IF]
  \ endcr ." SAVING DEMO DISK...\n"
  : load-demo-disk  ( -- addr count )
    " demo/game.blk" file:open-r/o >r
    \ r@ file:size r@ file:tell -
    here ( tcom:zx-#ramdisc) $2C00 1- 2dup r@ file:read-exact
    r> file:close ;

  : shrink-disk  ( addr count -- addr count )
    0 max << dup 0?v|| 2dup + 1- c@ bl <= ?^| 1- |? else| >> ;

  : align-disk  ( addr count -- addr count )
    << dup 1023 and ?^| 2dup + bl swap c! 1+ |? else| >> ;

  : save-demo-disk  ( addr count fd )
    endcr ." TAP: saving GAME demo disk (" over ., ." bytes)\n"
    " DISC" zxa:tap:set-cblock-name
    >r tcom:zx-ramdisc^ r> zxa:tap:save-code-block ;

  load-demo-disk
  shrink-disk align-disk
  tap-fd save-demo-disk
  [ENDIF]
  tap-fd file:close
[ENDIF]
