;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Z80 VMForth
;; Copyright (C) 2024 Ketmar Dark // Invisible Vector
;; Understanding is not required. Only obedience.
;; GPLv3 ONLY
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ZX control flow words
;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(*
<< cond [COR cond]* [CAND cond] ?^| ... |?
   cond [COR cond]* [CAND cond] ?v| ... |?
   else| ...
>>

i omited `init` predicate, because in Forth it is easy to factor it into
separate word (much easier than in other languages), and implementing
`init` efficiently requires extra branch at "<<". actually, several chained
branches, tracking of `init` in conditions, and such.

note that COR and CAND both doing "short-circuiting", jumping directly to
the corresponding "?.|" (or after the respective "|"). this nicely (and
efficiently) solves the problem with short-circuit boolean evaluation
without introducing a new control structure.

the last predicate should be either "else|", or non-conditional arrow ("^|"
or "v|"). the compiler will complain if you'll forget to do it.

you can use "^| ... |", or slightly more optimal "^||" as the last
predicate. of course, down arrow is allowed too. it is your responsbility
to not write any code after unconditional arrow.

one useful extension is "nested checks". it can be used to avoid
unnecessary stack manipulations, and carrying some flag around.

  ?<| some-code |>

this can be used instead of any "?.| ... |", and can be nested. this
"substructure" can contain futher checks, which repeat or exit the parent
"<< ... >>". this allows to factor the common part of several guard
expressions, evaluate it only once, and avoid passing it around (hence
eliminating extra stack manipulation).

there is no BREAK/CONTINUE for this control structure.

if your branches has no code, it is slightly more efficient to use "?^||"
and "^v||" words.

it is also possible to use "of?^|", "of?v|" and so on to implement "CASE".
it works as an ordinary "CASE" (i.e. drops the value on successfull check).

there is "?of?.|" construct. it expects the prepared condition. it can be
used to check ranges, or perform other non-standard tests:

  dup 1 3 within ?of?v| ... |?
  dup 5 > ?of?v| ... |?

note the "DUP" -- you have to keep the value you are testing. basically,
"n of?.|" is just an optimised "dup n = ?of?.|".

*)


<zx-definitions>
zx-here

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; compiler helpers

<zx-hidden>
<zx-system>

$c001 constant PAIR-IF
$c002 constant PAIR-IFELSE
$c003 constant PAIR-IFEXIT
;; Parnas' iterator pairs
$c010 constant PAIR-<<
$c011 constant PAIR-SHORT -- short-circuit boolean
$c012 constant PAIR-?<|
$c013 constant PAIR-?^|
$c014 constant PAIR-?v|
$c015 constant PAIR-^|
$c016 constant PAIR-v|


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; compiler helpers
;;

\ -1 constant of=brn
\ -2 constant of<>brn


;; if/else helper
: (IF-COMMON)     ( brnid -- chain pair-id )  SYS: ?comp SYS: CALL, SYS: mark> pair-if ;
: (IFEXIT-COMMON) ( brnid -- chain pair-id )  SYS: ?comp SYS: CALL, SYS: mark> pair-ifexit ;

\ : ?2PAIRS ( a b c )  >r over <> swap r> <> and ?error" unbalanced constructs" ;


\ module PARNAS-SUPPORT
\ <disable-hash>

;; Parnas' iterator helpers

0 quan PITTI-CONT
0 quan PITTI-BREAK
0 quan PITTI-NESTED
0 quan PITTI-ELSE?

|: PARNAS-SYS-RESET
  pitti-cont:!0 pitti-break:!0
  pitti-nested:!0 pitti-else?:!0 ;

: ?NO-ELSE  pitti-else? 11 ?error ;
: ?NO-NEST  pitti-nested 12 ?error ;

: PARNAS-NEST    pitti-nested:1+! ;
: PARNAS-UNNEST  pitti-nested:1-! ;
: PARNAS-ELSE    pitti-nested 0= pitti-else?:+! ;

: RESOLVE-PARNAS-SHORT  ( short-cont short-break brncfa -- short-break )
  dup ?< SYS: CALL, chain> || drop >? swap resolve> ;

;; this consumes the optional "short" info, and leaves the following:
;; ( short-break-chain over-jump-mark/0 newmark )
: PARNAS-BRANCH  ( prev-triplet brncfa newmark -- overbrn newmark )
  ?comp ?no-else >r over pair-short = ?< swap drop resolve-parnas-short
  || over pair-<< pair-?<| ?2pairs dup ?< SYS: CALL, mark> >? >? r> ;

: PARNAS-CONT   pitti-cont COMPILE SYS: BRANCH <resolve ;
: PARNAS-BREAK  pitti-break COMPILE SYS: BRANCH chain> pitti-break:! ;

: PARNAS-END  ( overbrn mark )
  dup pair-?^| pair-?v| ?2pairs
  pair-?^| = ?< parnas-cont || parnas-break >? resolve> ;

: INVERT-BRANCH  ( brn )
  dup ['] branch = ?exit
  ['] 0branch = ?< ['] tbranch || ['] 0branch >? ;

: PARNAS-SHORTINV  ( prev-triplet brncfa newmark -- overbrn newmark )
  swap invert-branch swap
  pair-?^| = ?< pitti-cont swap SYS: CALL, <resolve
  || pitti-break swap SYS: CALL, chain> pitti-break:! >? ;

;;TODO: optimise this!
: PARNAS-BRANCH-EMPTY  ( prev-triplet brncfa newmark -- overbrn newmark )
  ?comp ?no-else >r over pair-short = ?exit< r> parnas-branch parnas-end >?
  r> parnas-shortinv ;


;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; David Parnas' IT/TI flow control structure

<zx-forth>
<zx-normal>

*: <<
  ?comp pitti-cont pitti-break pitti-else? pitti-nested
  <mark
  pitti-cont:! pitti-break:!0 pitti-else?:!0 pitti-nested:!0
  pair-<< ;

*: >>
  ?comp pair-<< ?pairs ?no-nest
  pitti-else? 0= 10 ?error
  pitti-break resolve>
  pitti-nested:! pitti-else?:! pitti-break:! pitti-cont:! ;

;; COR/CAND info: cont-chain is after "?.|" or "<|", break-chain is after
*: COR  ?comp ?no-else
  dup pair-short = ?< rot COMPILE SYS: TBRANCH chain> nrot
  || dup pair-<< pair-?<| ?2pairs COMPILE SYS: TBRANCH mark> 0 pair-short >? ;

*: CAND  ?comp ?no-else
  dup pair-short = ?< swap COMPILE SYS: 0BRANCH chain> swap
  || dup pair-<< pair-?<| ?2pairs 0 COMPILE SYS: 0BRANCH mark> pair-short >? ;

*: |?  ( overbrn mark )  parnas-end ;

*: ?^|    ['] 0branch pair-?^| parnas-branch ;
*: ?v|    ['] 0branch pair-?v| parnas-branch ;
*: not?^| ['] tbranch pair-?^| parnas-branch ;
*: not?v| ['] tbranch pair-?v| parnas-branch ;
*: 0?^|   ['] tbranch pair-?^| parnas-branch ;
*: 0?v|   ['] tbranch pair-?v| parnas-branch ;
*: of?^|  ( of<>brn) pair-?^| COMPILE SYS: (OF<>) - ['] 0BRANCH parnas-branch COMPILE DROP ;
*: of?v|  ( of<>brn) pair-?v| COMPILE SYS: (OF<>) ['] 0BRANCH parnas-branch COMPILE DROP ;
*: <>of?^|  ( of=brn) pair-?^| COMPILE SYS: (OF<>) ['] TBRANCH parnas-branch COMPILE DROP ;
*: <>of?v|  ( of=brn) pair-?v| COMPILE SYS: (OF<>) ['] TBRANCH parnas-branch COMPILE DROP ;

1 [IF]
*: of?^||    [COMPILE] of?^| [COMPILE] |? ;
*: of?v||    [COMPILE] of?v| [COMPILE] |? ;
*: <>of?^||  [COMPILE] <>of?^| [COMPILE] |? ;
*: <>of?v||  [COMPILE] <>of?v| [COMPILE] |? ;
[ENDIF]

*: else|
  ?comp ?no-else ?no-nest
  dup pair-<< pair-?<| ?2pairs parnas-else ;

1 [IF]
*: ?^||    ['] 0branch pair-?^| parnas-branch-empty ;
*: ?v||    ['] 0branch pair-?v| parnas-branch-empty ;
*: not?^|| ['] tbranch pair-?^| parnas-branch-empty ;
*: not?v|| ['] tbranch pair-?v| parnas-branch-empty ;
*: 0?^||   [COMPILE] not?^|| ;
*: 0?v||   [COMPILE] not?v|| ;
[ENDIF]

*: ^|  dup pair-<< pair-?<| ?2pairs 0 pair-^| parnas-branch parnas-else ;
*: v|  dup pair-<< pair-?<| ?2pairs 0 pair-v| parnas-branch parnas-else ;

;; it is your responsitibility to NOT put any code after "|"
*: |  ( overbrn mark )
  dup pair-^| pair-v| ?2pairs
  pair-^| = ?< parnas-cont || pitti-nested ?< parnas-break >? >?
  resolve> ;

*: ^||  dup pair-<< pair-?<| ?2pairs parnas-cont parnas-else ;
*: v||  dup pair-<< pair-?<| ?2pairs pitti-nested ?< parnas-break >? parnas-else ;

1 [IF]
*: |>   ( overbrn mark )  pair-?<| ?pairs resolve> parnas-unnest ;
*: ?<|     dup pair-<< pair-short ?2pairs ['] 0branch pair-?<| parnas-branch parnas-nest ;
*: not?<|  dup pair-<< pair-short ?2pairs ['] tbranch pair-?<| parnas-branch parnas-nest ;
*: 0?<|    [COMPILE] not?<| ;
[ENDIF]

*: >?  ?comp dup pair-ifexit = ?< drop pair-if COMPILE EXIT >?
                pair-if pair-ifelse ?2pairs resolve> ;

*: ||  ?comp pair-if ?pairs COMPILE SYS: BRANCH mark>
                swap resolve> pair-ifelse ;

*: ?exit<     ['] 0branch (ifexit-common) ;
*: not?exit<  ['] tbranch (ifexit-common) ;
*: 0?exit<    [COMPILE] not?exit< ;

*: ?<     ['] 0branch (if-common) ;
*: not?<  ['] tbranch (if-common) ;
*: 0?<    [COMPILE] not?< ;


zx-here swap - ." Parnas' flowctl words size: " .bytes cr
<zx-done>
